from builtins import object
import logging
import json


# noinspection PyMethodMayBeStatic,PyUnusedLocal
class OnAWBase(object):
    """ Base class to be extended to extend ActingWeb functionality """

    def __init__(self):
        self.config = None
        self.myself = None
        self.webobj = None
        self.auth = None

    def aw_init(self, auth=None, webobj=None):
        self.config = auth.config
        self.myself = auth.actor
        self.webobj = webobj
        self.auth = auth

    def bot_post(self, path):
        """Called on POSTs to /bot.

        Note, there will not be any actor initialised.
        """

        # Safety valve to make sure we don't do anything if bot is not
        # configured.
        if not self.config.bot['token'] or len(self.config.bot['token']) == 0:
            return False
        if path == 'something':
            pass

        # try:
        #     body = json.loads(req.request.body.decode('utf-8', 'ignore'))
        #     logging.debug('Bot callback: ' + req.request.body.decode('utf-8', 'ignore'))
        # except:
        #     return 400
        #
        # This is how actor can be initialised if the bot request
        # contains a value that has been stored as an actor property.
        # This value must be a primary key for the external oauth identity
        # that the actor is representing.
        # Here, oauthId (from oauth service) has earlier been stored as a property
        # myself = actor.Actor()
        # myself.get_from_property(name='oauthId', value=<PROPERTY-VALUE>)
        # if myself.id:
        #    logging.debug('Found Actor(' + myself.id + ')')
        #
        # If we havent''
        # if not myself.id:
        #    myself.create(url=self.config.root, creator= <EMAIL>,
        #                    passphrase=self.config.new_token())
        # Now store the oauthId propery
        #    myself.store.oauthId = <ID-VALUE>
        # Send confirmation message that actor has been created
        #    return True
        # Do something
        return True

    def get_properties(self, path: list, data: dict) -> dict or None:
        """ Called on GET to properties for transformations to be done

        :param path: Target path requested
        :type path: list[str]
        :param data: Data retrieved from data store to be returned
        :type data: dict
        :return: The transformed data to return to requestor or None if 404 should be returned
        :rtype: dict or None
        """
        return data

    def delete_properties(self, path: list, old: dict, new: dict) -> bool:
        """ Called on DELETE to properties

        :param path: Target path to be deleted
        :type path: list[str]
        :param old: Property value that will be deleted (or changed)
        :type old: dict
        :param new: Property value after path has been deleted
        :type new: dict
        :return: True if DELETE is allowed, False if 403 should be returned
        :rtype: bool
        """
        return True

    def put_properties(self, path: list, old: dict, new: dict) -> dict or None:
        """ Called on PUT to properties for transformations to be done before save
        :param path: Target path requested to be updated
        :type path: list[str]
        :param old: Old data from database
        :type old: dict
        :param new:
        :type new: New data from PUT request (after merge)
        :return: The dict that should be stored or None if 400 should be returned and nothing stored
        :rtype: dict or None
        """
        return new

    def post_properties(self, prop: str, data: dict) -> dict or None:
        """ Called on POST to properties, once for each property

        :param prop: Property to be created
        :type prop: str
        :param data: The data to be stored in prop
        :type data: dict
        :return: The transformed data to store in prop or None if that property should be skipped and not stored
        :rtype: dict or None
        """
        return data

    def get_callbacks(self, name):
        """Customizible function to handle GET /callbacks"""
        # return True if callback has been processed
        # THE BELOW IS SAMPLE CODE
        # my_oauth=oauth.oauth(token = myself.store.oauth_token)
        if name == 'something':
            return
        # END OF SAMPLE CODE
        return False

    def delete_callbacks(self, name):
        """Customizible function to handle DELETE /callbacks"""
        # return True if callback has been processed
        return False

    def post_callbacks(self, name):
        """Customizible function to handle POST /callbacks"""
        # return True if callback has been processed
        # THE BELOW IS SAMPLE CODE
        # logging.debug("Callback body: "+req.request.body.decode('utf-8', 'ignore'))
        # non-json POSTs to be handled first
        # if name == 'somethingelse':
        #    return True
        # Handle json POSTs below
        # body = json.loads(req.request.body.decode('utf-8', 'ignore'))
        # data = body['data']
        # if name == 'somethingmore':
        #    callback_id = req.request.get('id')
        #    req.response.set_status(204)
        #    return True
        # req.response.set_status(403, "Callback not found.")
        # END OF SAMPLE CODE
        return False

    def post_subscriptions(self, sub, peerid, data):
        """Customizible function to process incoming callbacks/subscriptions/ callback with json body, return True if
        processed, False if not."""
        logging.debug("Got callback and processed " + sub["subscriptionid"] +
                      " subscription from peer " + peerid + " with json blob: " + json.dumps(data))
        return True

    def delete_actor(self):
        # THIS METHOD IS CALLED WHEN AN ACTOR IS REQUESTED TO BE DELETED.
        # THE BELOW IS SAMPLE CODE
        # Clean up anything associated with this actor before it is deleted.
        # END OF SAMPLE CODE
        return

    def check_on_oauth_success(self, token=None):
        # THIS METHOD IS CALLED WHEN AN OAUTH AUTHORIZATION HAS BEEN SUCCESSFULLY MADE AND BEFORE APPROVAL
        return True

    def actions_on_oauth_success(self):
        # THIS METHOD IS CALLED WHEN AN OAUTH AUTHORIZATION HAS BEEN SUCCESSFULLY MADE
        return True

    def get_resources(self, name):
        """ Called on GET to resources. Return struct for json out.

            Returning {} will give a 404 response back to requestor.
        """
        return {}

    def delete_resources(self, name):
        """ Called on DELETE to resources. Return struct for json out.

            Returning {} will give a 404 response back to requestor.
        """
        return {}

    def put_resources(self, name, params):
        """ Called on PUT to resources. Return struct for json out.

            Returning {} will give a 404 response back to requestor.
            Returning an error code after setting the response will not change
            the error code.
        """
        return {}

    def post_resources(self, name, params):
        """ Called on POST to resources. Return struct for json out.

            Returning {} will give a 404 response back to requestor.
            Returning an error code after setting the response will not change
            the error code.
        """
        return {}

    def www_paths(self, path=''):
        # THIS METHOD IS CALLED WHEN AN actorid/www/* PATH IS CALLED (AND AFTER ACTINGWEB DEFAULT PATHS HAVE BEEN
        # HANDLED)
        # THE BELOW IS SAMPLE CODE
        # if path == '' or not myself:
        #    logging.info('Got an on_www_paths without proper parameters.')
        #    return False
        # if path == 'something':
        #    return True
        # END OF SAMPLE CODE
        return False



from builtins import object
import os
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection
import logging

"""
    DbActor handles all db operations for an actor
    Google datastore for google is used as a backend.
"""


class CreatorIndex(GlobalSecondaryIndex):
    """
    Secondary index on actor
    """
    class Meta(object):
        index_name = 'creator-index'
        read_capacity_units = 2
        write_capacity_units = 1
        projection = AllProjection()

    creator = UnicodeAttribute(hash_key=True)


class Actor(Model):
    """
       DynamoDB data model for an actor
    """
    class Meta(object):
        table_name = os.getenv('AWS_DB_PREFIX', 'demo_actingweb') + "_actors"
        read_capacity_units = 6
        write_capacity_units = 2
        region = os.getenv('AWS_DEFAULT_REGION', 'us-west-1')
        host = os.getenv('AWS_DB_HOST', None)

    id = UnicodeAttribute(hash_key=True)
    creator = UnicodeAttribute()
    passphrase = UnicodeAttribute()
    creator_index = CreatorIndex()


class DbActor(object):
    """
        DbActor does all the db operations for actor objects

    """

    def get(self,  actor_id=None):
        """ Retrieves the actor from the database """
        if not actor_id:
            return None
        try:
            self.handle = Actor.get(actor_id, consistent_read=True)
        except Actor.DoesNotExist:
            return None
        if self.handle:
            t = self.handle
            return {
                "id": t.id,
                "creator": t.creator,
                "passphrase": t.passphrase,
            }
        else:
            return None

    def get_by_creator(self, creator=None):
        """ Retrieves the actor from db based on creator field

            Returns None if none was found. If one is found, that one is
            loaded in the object. If more, all are returned.
        """
        if not creator:
            return None
        if '@' in creator:
            creator = creator.lower()
        self.handle = Actor.creator_index.query(creator)
        ret = []
        for c in self.handle:
            logging.warning("    id (" + c.id + ")")
            ret.append(self.get(actor_id=c.id))
        return ret

    def modify(self, creator=None, passphrase=None):
        """ Modify an actor """
        if not self.handle:
            logging.debug("Attempted modification of DbActor without db handle")
            return False
        if creator and len(creator) > 0:
            # Email in creator needs to be lower case
            if '@' in creator:
                creator = creator.lower()
            self.handle.creator = creator
        if passphrase and len(passphrase) > 0:
            self.handle.passphrase = passphrase.decode('utf-8')
        self.handle.save()
        return True

    def create(self, actor_id=None, creator=None,
               passphrase=None):
        """ Create a new actor """
        if not actor_id:
            return False
        if not creator:
            creator = ''
        # Email in creator needs to be lower case
        if '@' in creator:
            creator = creator.lower()
        if not passphrase:
            passphrase = ''
        if self.get(actor_id=actor_id):
            logging.warning("Trying to create actor that exists(" + actor_id + ")")
            return False
        self.handle = Actor(id=actor_id,
                            creator=creator,
                            passphrase=passphrase)
        self.handle.save()
        return True

    def delete(self):
        """ Deletes the actor in the database """
        if not self.handle:
            logging.debug("Attempted delete of DbActor without db handle")
            return False
        self.handle.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        if not Actor.exists():
            Actor.create_table(wait=True)


class DbActorList(object):
    """
        DbActorList does all the db operations for list of actor objects
    """

    def fetch(self):
        """ Retrieves the actors in the database """
        self.handle = Actor.scan()
        if self.handle:
            ret = []
            for t in self.handle:
                ret.append({
                    "id": t.id,
                    "creator": t.creator,
                    })
            return ret
        else:
            return False

    def __init__(self):
        self.handle = None



from builtins import str
import json
import logging

from actingweb import auth
from actingweb.handlers import base_handler


class CallbacksHandler(base_handler.BaseHandler):

    def get(self, actor_id, name):
        """Handles GETs to callbacks"""
        if self.request.get('_method') == 'PUT':
            self.put(actor_id, name)
        if self.request.get('_method') == 'POST':
            self.post(actor_id, name)
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='callbacks',
                                              add_response=False,
                                              config=self.config
                                              )
        if not myself or (check.response["code"] != 200 and check.response["code"] != 401):
            auth.add_auth_response(appreq=self, auth_obj=check)
            return
        if not check.check_authorisation(path='callbacks', subpath=name, method='GET'):
            self.response.set_status(403, 'Forbidden')
            return
        if not self.on_aw.get_callbacks(name=name):
            self.response.set_status(403, 'Forbidden')

    def put(self, actor_id, name):
        """PUT requests are handled as POST for callbacks"""
        self.post(actor_id, name)

    def delete(self, actor_id, name):
        """Handles deletion of callbacks, like subscriptions"""
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='callbacks',
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        path = name.split('/')
        if path[0] == 'subscriptions':
            peerid = path[1]
            subid = path[2]
            if not check.check_authorisation(path='callbacks', subpath='subscriptions', method='DELETE', peerid=peerid):
                self.response.set_status(403, 'Forbidden')
                return
            sub = myself.get_subscription_obj(peerid=peerid, subid=subid, callback=True)
            if sub:
                sub.delete()
                self.response.set_status(204, 'Deleted')
                return
            self.response.set_status(404, 'Not found')
            return
        if not check.check_authorisation(path='callbacks', subpath=name, method='DELETE'):
            self.response.set_status(403, 'Forbidden')
            return
        if not self.on_aw.delete_callbacks(name=name):
            self.response.set_status(403, 'Forbidden')

    def post(self, actor_id, name):
        """Handles POST callbacks"""
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='callbacks',
                                              add_response=False,
                                              config=self.config)
        # Allow unauthenticated requests to /callbacks/subscriptions, so
        # do the auth check further below
        path = name.split('/')
        if path[0] == 'subscriptions':
            peerid = path[1]
            subid = path[2]
            sub = myself.get_subscription(peerid=peerid, subid=subid, callback=True)
            if sub and len(sub) > 0:
                logging.debug("Found subscription (" + str(sub) + ")")
                if not check.check_authorisation(
                        path='callbacks',
                        subpath='subscriptions',
                        method='POST',
                        peerid=peerid):
                    self.response.set_status(403, 'Forbidden')
                    return
                try:
                    params = json.loads(self.request.body.decode('utf-8', 'ignore'))
                except (TypeError, ValueError, KeyError):
                    self.response.set_status(400, "Error in json body")
                    return
                if self.on_aw.post_subscriptions(sub=sub, peerid=peerid, data=params):
                    self.response.set_status(204, 'Found')
                else:
                    self.response.set_status(400, 'Processing error')
                return
            self.response.set_status(404, 'Not found')
            return
        if not myself or (check.response["code"] != 200 and check.response["code"] != 401):
            auth.add_auth_response(appreq=self, auth_obj=check)
            return
        if not check.check_authorisation(path='callbacks', subpath=name, method='POST'):
            self.response.set_status(403, 'Forbidden')
            return
        if not self.on_aw.post_callbacks(name=name):
            self.response.set_status(403, 'Forbidden')



from builtins import object
import logging


class PeerTrustee(object):

    def get(self):
        if self.peertrustee and len(self.peertrustee) > 0:
            return self.peertrustee
        self.peertrustee = self.handle.get(actor_id=self.actor_id,
                                           peerid=self.peerid,
                                           peer_type=self.peer_type)
        return self.peertrustee

    def create(self, baseuri=None, passphrase=None):
        if not self.handle:
            self.handle = self.config.DbPeerTrustee.DbPeerTrustee()
        if not self.actor_id or not self.peerid:
            logging.debug("Attempt to create new peer trustee without actor_id or peerid set")
            return False
        if not self.peer_type or len(self.peer_type) == 0:
            logging.debug("Attempt to create peer trustee without peer_type set.")
            return False
        return self.handle.create(actor_id=self.actor_id,
                                  peerid=self.peerid,
                                  peer_type=self.peer_type,
                                  baseuri=baseuri,
                                  passphrase=passphrase)

    def delete(self):
        if not self.handle:
            logging.debug("Attempt to delete peertrustee without db handle")
            return False
        return self.handle.delete()

    def __init__(self, actor_id=None, peerid=None, short_type=None, peer_type=None, config=None):
        self.config = config
        self.handle = self.config.DbPeerTrustee.DbPeerTrustee()
        self.peertrustee = {}
        self.peer_type = None
        if not actor_id or len(actor_id) == 0:
            logging.debug("No actorid set in initialisation of peertrust")
            return
        if peer_type:
            self.peer_type = peer_type
        elif not peer_type and short_type:
            if not self.config.actors[short_type]:
                logging.error('Got request to initialise peer trustee with unknown shortpeer_type(' + peer_type +
                              ')')
                return
            self.peer_type = self.config.actors[short_type]["type"]
        elif not peerid:
            logging.debug("Peerid and short_type are not set in initialisation of peertrustee. One is required")
            return
        self.actor_id = actor_id
        self.peerid = peerid
        self.get()



from builtins import object
import datetime
import logging


class Subscription(object):
    """Base class with core subscription methods (storage-related)"""

    def get(self):
        """Retrieve subscription from db given pre-initialized variables """
        if not self.actor_id or not self.peerid or not self.subid:
            return None
        if self.subscription and len(self.subscription) > 0:
            return self.subscription
        self.subscription = self.handle.get(actor_id=self.actor_id,
                                            peerid=self.peerid,
                                            subid=self.subid)
        if not self.subscription:
            self.subscription = {}
        return self.subscription

    def create(self, target=None, subtarget=None, resource=None, granularity=None, seqnr=1):
        """Create new subscription and push it to db"""
        if self.subscription and len(self.subscription) > 0:
            logging.debug("Attempted creation of subscription when already loaded from storage")
            return False
        if not self.actor_id or not self.peerid:
            logging.debug("Attempted creation of subscription without actor_id or peerid set")
            return False
        if not self.subid:
            now = datetime.datetime.utcnow()
            seed = self.config.root + now.strftime("%Y%m%dT%H%M%S%f")
            self.subid = self.config.new_uuid(seed)
        if not self.handle.create(actor_id=self.actor_id,
                                  peerid=self.peerid,
                                  subid=self.subid,
                                  granularity=granularity,
                                  target=target,
                                  subtarget=subtarget,
                                  resource=resource,
                                  seqnr=seqnr,
                                  callback=self.callback):
            return False
        self.subscription["id"] = self.actor_id
        self.subscription["subscriptionid"] = self.subid
        self.subscription["peerid"] = self.peerid
        self.subscription["target"] = target
        self.subscription["subtarget"] = subtarget
        self.subscription["resource"] = resource
        self.subscription["granularity"] = granularity
        self.subscription["sequence"] = seqnr
        self.subscription["callback"] = self.callback
        return True

    def delete(self):
        """Delete a subscription in storage"""
        if not self.handle:
            logging.debug("Attempted delete of subscription without storage handle")
            return False
        self.clear_diffs()
        self.handle.delete()
        return True

    def increase_seq(self):
        if not self.handle:
            logging.debug("Attempted increase_seq without subscription retrieved from storage")
            return False
        self.subscription["sequence"] += 1
        return self.handle.modify(seqnr=self.subscription["sequence"])

    def add_diff(self, blob=None):
        """Add a new diff for this subscription"""
        if not self.actor_id or not self.subid or not blob:
            logging.debug("Attempted add_diff without actorid, subid, or blob")
            return False
        diff = self.config.DbSubscriptionDiff.DbSubscriptionDiff()
        diff.create(actor_id=self.actor_id,
                    subid=self.subid,
                    diff=blob,
                    seqnr=self.subscription["sequence"]
                    )
        if not self.increase_seq():
            logging.error("Failed increasing sequence number for subscription " +
                          self.subid + " for peer " + self.peerid)
        return diff.get()

    def get_diff(self, seqnr=0):
        """Get one specific diff"""
        if seqnr == 0:
            return None
        if not isinstance(seqnr, int):
            return None
        diff = self.config.DbSubscriptionDiff.DbSubscriptionDiff()
        return diff.get(actor_id=self.actor_id, subid=self.subid, seqnr=seqnr)

    def get_diffs(self):
        """Get all the diffs available for this subscription ordered by the timestamp, oldest first"""
        diff_list = self.config.DbSubscriptionDiff.DbSubscriptionDiffList()
        return diff_list.fetch(actor_id=self.actor_id, subid=self.subid)

    def clear_diff(self, seqnr):
        """Clears one specific diff"""
        diff = self.config.DbSubscriptionDiff.DbSubscriptionDiff()
        diff.get(actor_id=self.actor_id, subid=self.subid, seqnr=seqnr)
        return diff.delete()

    def clear_diffs(self, seqnr=0):
        """Clear all diffs up to and including a seqnr"""
        diff_list = self.config.DbSubscriptionDiff.DbSubscriptionDiffList()
        diff_list.fetch(actor_id=self.actor_id, subid=self.subid)
        diff_list.delete(seqnr=seqnr)

    def __init__(self, actor_id=None, peerid=None, subid=None, callback=False, config=None):
        self.config = config
        self.handle = self.config.DbSubscription.DbSubscription()
        self.subscription = {}
        if not actor_id:
            return
        self.actor_id = actor_id
        self.peerid = peerid
        self.subid = subid
        self.callback = callback
        if self.actor_id and self.peerid and self.subid:
            self.get()


class Subscriptions(object):
    """ Handles all subscriptions of a specific actor_id

        Access the indvidual subscriptions in .dbsubscriptions and the subscription data
        in .subscriptions as a dictionary
    """

    def fetch(self):
        if self.subscriptions is not None:
            return self.subscriptions
        if not self.list:
            self.list = self.config.DbSubscription.DbSubscriptionList()
        if not self.subscriptions:
            self.subscriptions = self.list.fetch(actor_id=self.actor_id)
        return self.subscriptions

    def delete(self):
        if not self.list:
            logging.debug("Already deleted list in subscriptions")
            return False
        for sub in self.subscriptions:
            diff_list = self.config.DbSubscriptionDiff.DbSubscriptionDiffList()
            diff_list.fetch(actor_id=self.actor_id, subid=sub["subscriptionid"])
            diff_list.delete()
        self.list.delete()
        self.list = None
        self.subscriptions = None
        return True

    def __init__(self,  actor_id=None, config=None):
        """ Properties must always be initialised with an actor_id """
        self.config = config
        if not actor_id:
            self.list = None
            logging.debug("No actor_id in initialisation of subscriptions")
            return
        self.list = self.config.DbSubscription.DbSubscriptionList()
        self.actor_id = actor_id
        self.subscriptions = None
        self.fetch()



from builtins import str
from builtins import object
try:
    from urllib.parse import urlencode as urllib_urlencode
except ImportError:
    from urllib import urlencode as urllib_urlencode
import logging
import json
import base64


# This function code is from latest urlfetch. For some reason the
# Appengine version of urlfetch does not include links()


def pagination_links(self):
    """Links parsed from HTTP Link header"""
    rets = []
    if 'link' in self.headers:
        linkheader = self.headers['link']
    else:
        return rets
    for i in linkheader.split(','):
        try:
            url, params = i.split(';', 1)
        except ValueError:
            url, params = i, ''
        link = {'url': url.strip('''<> '"''')}
        for param in params.split(';'):
            try:
                k, v = param.split('=')
            except ValueError:
                break
            link[k.strip(''' '"''')] = v.strip(''' '"''')
        rets.append(link)
    return rets


class OAuth(object):

    def __init__(self, token=None, config=None):
        self.config = config
        self.token = token
        self.first = None
        self.next = None
        self.prev = None
        self.last_response_code = 0
        self.last_response_message = ""

    def enabled(self):
        if len(self.config.oauth['client_id']) == 0:
            return False
        else:
            return True

    def set_token(self, token):
        if token:
            self.token = token

    def post_request(self, url, params=None, urlencode=False, basic_auth=False):
        if params:
            if urlencode:
                data = urllib_urlencode(params)
                logging.debug('Oauth POST request with urlencoded payload: ' + url + ' ' + data)
            else:
                data = json.dumps(params)
                logging.debug('Oauth POST request with JSON payload: ' + url + ' ' + data)
        else:
            data = None
            logging.debug('Oauth POST request: ' + url)
        if urlencode:
            if self.token:
                if basic_auth:
                    headers = {'Content-Type': 'application/x-www-form-urlencoded',
                            'Authorization': 'Basic ' + base64.b64encode(self.token.encode("utf-8")).decode("utf-8"),
                            }
                else:
                    headers = {'Content-Type': 'application/x-www-form-urlencoded',
                            'Authorization': 'Bearer ' + self.token,
                            }
            else:
                headers = {'Content-Type': 'application/x-www-form-urlencoded',
                           }
        else:
            if self.token:
                headers = {'Content-Type': 'application/json',
                           'Authorization': 'Bearer ' + self.token,
                           }
            else:
                headers = {'Content-Type': 'application/json'}
        try:
            if self.config.env == 'appengine':
                self.config.module["urlfetch"].set_default_fetch_deadline(20)
                response = self.config.module["urlfetch"].fetch(
                    url=url,
                    payload=data,
                    method=self.config.module["urlfetch"].POST,
                    headers=headers)
            else:
                response = self.config.module["urlfetch"].post(url=url, data=data, headers=headers)
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            self.last_response_code = 0
            self.last_response_message = 'No response'
            logging.warning("Oauth POST failed with exception")
            return None
        if response.status_code == 204:
            return {}
        if response.status_code != 200 and response.status_code != 201:
            logging.info('Error when sending POST request: ' +
                         str(response.status_code) + str(response.content))
            return None
        logging.debug('Oauth POST response JSON:' + str(response.content))
        return json.loads(response.content.decode('utf-8', 'ignore'))

    def put_request(self, url, params=None, urlencode=False):
        if params:
            if urlencode:
                data = urllib_urlencode(params)
                logging.debug('Oauth PUT request with urlencoded payload: ' + url + ' ' + data)
            else:
                data = json.dumps(params)
                logging.debug('Oauth PUT request with JSON payload: ' + url + ' ' + data)
        else:
            data = None
            logging.debug('Oauth PUT request: ' + url)
        if urlencode:
            if self.token:
                headers = {'Content-Type': 'application/x-www-form-urlencoded',
                           'Authorization': 'Bearer ' + self.token,
                           }
            else:
                headers = {'Content-Type': 'application/x-www-form-urlencoded',
                           }
        else:
            if self.token:
                headers = {'Content-Type': 'application/json',
                           'Authorization': 'Bearer ' + self.token,
                           }
            else:
                headers = {'Content-Type': 'application/json'}
        try:
            if self.config.env == 'appengine':
                self.config.module["urlfetch"].set_default_fetch_deadline(20)
                response = self.config.module["urlfetch"](
                    url=url,
                    payload=data,
                    method=self.config.module["urlfetch"].PUT,
                    headers=headers)
            else:
                response = self.config.module["urlfetch"].put(url=url, data=data, headers=headers)
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            self.last_response_code = 0
            self.last_response_message = 'No response'
            logging.warning("Oauth PUT failed with exception")
            return None
        if response.status_code == 204:
            return {}
        if response.status_code != 200 and response.status_code != 201:
            logging.info('Error when sending PUT request: ' +
                         str(response.status_code) + str(response.content))
            return None
        logging.debug('Oauth PUT response JSON:' + str(response.content))
        return json.loads(response.content.decode('utf-8', 'ignore'))

    def get_request(self, url, params=None):
        if not self.token:
            logging.debug("No token set in get_request()")
            return None
        if params:
            url = url + '?' + urllib_urlencode(params)
        logging.debug('Oauth GET request: ' + url)
        try:
            if self.config.env == 'appengine':
                self.config.module["urlfetch"].set_default_fetch_deadline(60)
                response = self.config.module["urlfetch"].fetch(
                    url=url,
                    method=self.config.module["urlfetch"].GET,
                    headers={
                        'Authorization': 'Bearer ' + self.token,
                    }
                    )
            else:
                response = self.config.module["urlfetch"].get(
                    url,
                    headers={
                        'Authorization': 'Bearer ' + self.token,
                    }
                    )
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            self.last_response_code = 0
            self.last_response_message = 'No response'
            logging.warning("Oauth GET failed with exception")
            return None
        if response.status_code < 200 or response.status_code > 299:
            logging.info('Error when sending GET request to Oauth: ' +
                         str(response.status_code) + str(response.content))
            return None
        links = pagination_links(response)
        self.next = None
        self.first = None
        self.prev = None
        for link in links:
            logging.debug('Links:' + link['rel'] + ':' + link['url'])
            if link['rel'] == 'next':
                self.next = link['url']
            elif link['rel'] == 'first':
                self.first = link['url']
            elif link['rel'] == 'prev':
                self.prev = link['url']
        return json.loads(response.content.decode('utf-8', 'ignore'))

    def head_request(self, url, params=None):
        if not self.token:
            logging.debug("No token set in head_request(()")
            return None
        if params:
            url = url + '?' + urllib_urlencode(params)
        logging.debug('Oauth HEAD request: ' + url)
        try:
            response = self.config.module["urlfetch"].head(
                    url=url,
                    headers={'Authorization': 'Bearer ' + self.token}
                    )
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            self.last_response_code = 0
            self.last_response_message = 'No response'
            logging.warning("Oauth HEAD failed with exception")
            return None
        if response.status_code < 200 or response.status_code > 299:
            logging.warning('Error when sending HEAD request to Oauth: ' +
                            str(response.status_code) + str(response.content))
            return None
        return response.headers

    def delete_request(self, url):
        if not self.token:
            return None
        logging.debug('Oauth DELETE request: ' + url)
        try:
            if self.config.env == 'appengine':
                response = self.config.module["urlfetch"].fetch(url=url,
                                                                method=self.config.module["urlfetch"].DELETE,
                                                                headers={'Authorization': 'Bearer ' + self.token}
                                                                )
            else:
                response = self.config.module["urlfetch"].delete(
                    url=url,
                    headers={'Authorization': 'Bearer ' + self.token}
                    )
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            logging.warning("Spark DELETE failed.")
            self.last_response_code = 0
            self.last_response_message = 'No response'
            return None
        if response.status_code < 200 or response.status_code > 299:
            logging.info('Error when sending DELETE request to Oauth: ' +
                         str(response.status_code) + str(response.content))
            return None
        if response.status_code == 204:
            return {}
        try:
            ret = json.loads(response.content.decode('utf-8', 'ignore'))
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            return {}
        return ret

    def oauth_redirect_uri(self, state='', creator=None):
        params = {
            'response_type': self.config.oauth['response_type'],
            'client_id': self.config.oauth['client_id'],
            'redirect_uri': self.config.oauth['redirect_uri'],
            'scope': self.config.oauth['scope'],
            'state': state,
        }
        if 'oauth_extras' in self.config.oauth:
            for k, v in self.config.oauth['oauth_extras'].items():
                if isinstance(v, str) and 'dynamic:' in v:
                    if v[8:] == 'creator' and creator:
                        v = creator
                params[k] = v
        uri = self.config.oauth['auth_uri'] + "?" + urllib_urlencode(params)
        logging.debug('OAuth redirect with url: ' + uri + ' and state:' + state)
        return uri

    def oauth_request_token(self, code=None):
        if not code:
            return None
        params = {
            'grant_type': self.config.oauth['grant_type'],
            'client_id': self.config.oauth['client_id'],
            'client_secret': self.config.oauth['client_secret'],
            'code': code,
            'redirect_uri': self.config.oauth['redirect_uri'],
        }
        # Some OAuth2 implementations require Basic auth with client_id and secret
        self.token = self.config.oauth['client_id'] + ':' + self.config.oauth['client_secret']
        result = self.post_request(url=self.config.oauth[
                                  'token_uri'], params=params, urlencode=True, basic_auth=True)
        if result and 'access_token' in result:
            self.token = result['access_token']
        return result

    def oauth_refresh_token(self, refresh_token):
        if not refresh_token:
            return None
        params = {
            'grant_type': self.config.oauth['refresh_type'],
            'client_id': self.config.oauth['client_id'],
            'client_secret': self.config.oauth['client_secret'],
            'refresh_token': refresh_token,
        }
        # Some OAuth2 implementations require Basic auth with client_id and secret
        self.token = self.config.oauth['client_id'] + ':' + self.config.oauth['client_secret']
        result = self.post_request(url=self.config.oauth[
                                  'token_uri'], params=params, urlencode=True, basic_auth=True)
        if not result:
            self.token = None
            return None
        self.token = result['access_token']
        return result



import json
from actingweb import auth
from actingweb.handlers import base_handler


class MetaHandler(base_handler.BaseHandler):

    def get(self, actor_id, path):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id, path='meta',
                                              subpath=path,
                                              add_response=False,
                                              config=self.config)
        # We accept no auth here, so don't check response code
        if not myself:
            return
        if not check.check_authorisation(path='meta', subpath=path, method='GET', approved=False):
            self.response.set_status(403)
            return

        trustee_root = myself.store.trustee_root
        if self.config.migrate_2_5_0 and not trustee_root:
            trustee_root = myself.property.trustee_root
            if trustee_root:
                myself.property.trustee_root = None
                myself.store.trustee_root = trustee_root
        if not trustee_root:
            trustee_root = ''
        if not path:
            values = {
                'id': actor_id,
                'type': self.config.aw_type,
                'version': self.config.version,
                'desc': self.config.desc,
                'info': self.config.info,
                'trustee_root': trustee_root,
                'specification': self.config.specification,
                'aw_version': self.config.aw_version,
                'aw_supported': self.config.aw_supported,
                'aw_formats': self.config.aw_formats,
            }
            out = json.dumps(values)
            self.response.write(out.encode('utf-8'))
            self.response.headers["Content-Type"] = "application/json"
            return

        elif path == 'id':
            out = actor_id
        elif path == 'type':
            out = self.config.aw_type
        elif path == 'version':
            out = self.config.version
        elif path == 'desc':
            out = self.config.desc + myself.id
        elif path == 'info':
            out = self.config.info
        elif path == 'trustee_root':
            out = trustee_root
        elif path == 'specification':
            out = self.config.specification
        elif path == 'actingweb/version':
            out = self.config.aw_version
        elif path == 'actingweb/supported':
            out = self.config.aw_supported
        elif path == 'actingweb/formats':
            out = self.config.aw_formats
        else:
            self.response.set_status(404)
            return
        self.response.write(out.encode('utf-8'))



from actingweb import auth
from actingweb.handlers import base_handler


class WwwHandler(base_handler.BaseHandler):

    def get(self, actor_id, path):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id, path='www', subpath=path,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not self.config.ui:
            self.response.set_status(404, "Web interface is not enabled")
            return
        if not check.check_authorisation(path='www', subpath=path, method='GET'):
            self.response.write('')
            self.response.set_status(403)
            return

        if not path or path == '':
            self.response.template_values = {
                'url': self.request.url,
                'id': actor_id,
                'creator': myself.creator,
                'passphrase': myself.passphrase,
            }
            return

        if path == 'init':
            self.response.template_values = {
                'id': myself.id,
            }
            return
        if path == 'properties':
            properties = myself.get_properties()
            properties = self.on_aw.get_properties(path=None, data=properties)
            self.response.template_values = {
                'id': myself.id,
                'properties': properties,
            }
            return
        if path == 'property':
            lookup = myself.property[self.request.get('name')]
            lookup = self.on_aw.get_properties(
                path=self.request.get('name'), data=lookup)
            if lookup:
                self.response.template_values = {
                    'id': myself.id,
                    'property': self.request.get('name'),
                    'value': lookup,
                    'qual': '',
                }
            else:
                self.response.template_values = {
                    'id': myself.id,
                    'property': self.request.get('name'),
                    'value': 'Not set',
                    'qual': 'no',
                }
            return
        if path == 'trust':
            relationships = myself.get_trust_relationships()
            if not relationships or len(relationships) == 0:
                self.response.set_status(404, 'Not found')
                return
            for t in relationships:
                t["approveuri"] = self.config.root + myself.id + \
                    '/trust/' + t.relationship + '/' + t.peerid
                self.response.template_values = {
                    'id': myself.id,
                    'trusts': relationships,
                }
            return
        output = self.on_aw.www_paths(path=path)
        if output:
            self.response.write(output)
        else:
            self.response.set_status(404, "Not found")
        return



from __future__ import absolute_import
from builtins import str
from builtins import object
import logging
import time
import base64
import math
from actingweb import (actor, oauth, trust, config as config_class)

# This is where each path and subpath in actingweb is assigned an authentication type
# Fairly simple: /oauth is always oauth, /www can be either basic+trust or
# oauth through config.py, and everything else is basic+trust


def select_auth_type(path, subpath, config=None):
    """Selects authentication type based on path and subpath.

    Currently are only basic and oauth supported. Peer auth is automatic if an Authorization Bearer <token> header is
    included in the http request.
    """
    if path == 'oauth':
        return 'oauth'
    if path == 'www':
        return config.www_auth
    if subpath == 'oauth':
        return 'oauth'
    return 'basic'


def add_auth_response(appreq=None, auth_obj=None):
    """Called after init_actingweb() if add_response was set to False, and now responses should be added.
    """
    if not appreq or not auth_obj:
        return False
    logging.debug("add_auth_response: " +
                  str(auth_obj.response['code']) + ":" + auth_obj.response['text'])
    appreq.set_status(auth_obj.response['code'], auth_obj.response['text'])
    if auth_obj.response['code'] == 302:
        appreq.set_redirect(url=auth_obj.redirect)
    elif auth_obj.response['code'] == 401:
        appreq.write("Authentication required")
    for h, v in list(auth_obj.response['headers'].items()):
        appreq.headers[h] = v
    return True


def init_actingweb(appreq=None, actor_id=None, path='', subpath='', add_response=True, config=None):
    """Initialises actingweb by loading a config object, an actor object, and authentication object.


            More details about the authentication can be found in the auth object. If add_response is True,
            appreq.response
            will be changed according to authentication result. If False, appreq will not be touched.
            authn_done (bool), response['code'], response['text'], and response['headers'] will indicate results of
            the authentication process
            and need to be acted upon.
            200 is access approved
            302 is redirect and auth_obj.redirect contains redirect location where response must be redirected
            401 is authentication required, response['headers'] must be added to response
            403 is forbidden, text in response['text']
    """

    fullpath = '/' + path + '/' + subpath
    auth_type = select_auth_type(path=path, subpath=subpath, config=config)
    auth_obj = Auth(actor_id, auth_type=auth_type, config=config)
    if not auth_obj.actor:
        if add_response:
            appreq.response.set_status(404, 'Actor not found')
        return None, None
    auth_obj.check_authentication(appreq=appreq, path=fullpath)
    if add_response:
        add_auth_response(appreq.response, auth_obj)
    # Initialize the on_aw object with auth (.actor) and webobj (.config) for functions in on_aw
    appreq.on_aw.aw_init(auth=auth_obj, webobj=appreq)
    return auth_obj.actor, auth_obj


class Auth(object):
    """ The auth class handles authentication and authorisation for the various schemes supported.

    The helper function init_actingweb() can be used to give you an auth object and do authentication (or can be called
    directly).
    The check_authentication() function checks the various authentication schemes against the path and does proper
    authentication.
    There are three types supported: basic (using creator credentials), token (received when trust is created), or
    oauth (used to bind
    an actor to an oauth-enabled external service, as well as to log into /www path where interactive web functionality
    of the actor is available).
    The check_authorisation() function validates the authenticated user against the config.py access list.
    check_token_auth() can be called from outside the class to do a simple peer/bearer token verification.
    The OAuth helper functions are used to:
    process_oauth_callback() - process an OAuth callback as part of an OAuth flow and exchange code with a valid token
    validate_oauth_token() - validate and, if necessary, refresh a token
    set_cookie_on_cookie_redirect() - set a session cookie in the browser to the token value (called AFTER OAuth has
    been done!)

    The response[], acl[], and authn_done variables are useful outside Auth(). authn_done is set when authentication has
    been done and a final authentication status can be found in response[].

         self.response = {

            "code": 403,                # Result code (http)
            "text": "Forbidden",        # Proposed response text
            "headers": [],              # Headers to add to response after authentication has been done

        }

        self.acl = {

            "authenticated": False, # Has authentication been verified and passed?
            "authorised": False,    # Has authorisation been done and appropriate acls set?
            "rights": '',           # "a", "r" (approve or reject)
            "relationship": None,   # E.g. creator, friend, admin, etc
            "peerid": '',           # Peerid if there is a relationship
            "approved": False,      # True if the peer is approved

        }

    """

    def __init__(self, actor_id, auth_type='basic', config=None):
        if not config:
            self.config = config_class.Config()
        else:
            self.config = config
        self.token = None
        self.cookie_redirect = None
        self.cookie = None
        self.type = auth_type
        self.trust = None
        self.oauth = None
        # Proposed response code after check_authentication() or authorise() have been called
        self.response = {'code': 403, 'text': "Forbidden", 'headers': {}}
        # Whether authentication is complete or not (depends on flow)
        self.authn_done = False
        # acl stores the actual verified credentials and access rights after
        # authentication and authorisation have been done
        self.acl = {
            "authenticated": False,  # Has authentication been verified and passed?
            "authorised": False,    # Has authorisation been done and appropriate acls set?
            "rights": '',           # "a", "r" (approve or reject)
            "relationship": None,   # E.g. creator, friend, admin, etc
            "peerid": '',           # Peerid if there is a relationship
            "approved": False,      # True if the peer is approved
        }
        self.actor = actor.Actor(actor_id, config=self.config)
        if not self.actor.id:
            self.actor = None
            self.oauth = oauth.OAuth(token=None, config=self.config)
            self.token = None
            self.expiry = None
            self.refresh_expiry = None
            self.refresh_token = None
            return
        # We need to initialise oauth for use towards the external oauth service
        # Property name used to set self.token
        self.oauth_token_property = 'oauth_token'
        self.token = self.actor.store.oauth_token
        if self.config.migrate_2_5_0 and not self.token:
            self.token = self.actor.property.oauth_token
            if self.token:
                self.actor.store.oauth_token = self.token
                self.actor.property.oauth_token = None
        self.oauth = oauth.OAuth(token=self.token, config=self.config)
        self.expiry = self.actor.store.oauth_token_expiry
        self.refresh_expiry = self.actor.store.oauth_refresh_token_expiry
        self.refresh_token = self.actor.store.oauth_refresh_token
        if self.config.migrate_2_5_0:
            if not self.expiry:
                self.expiry = self.actor.property.oauth_token_expiry
                if self.expiry:
                    self.actor.store.oauth_token_expiry = self.expiry
                    self.actor.property.oauth_token_expiry = None
            if not self.refresh_expiry:
                self.refresh_expiry = self.actor.property.oauth_refresh_token_expiry
                if self.refresh_expiry:
                    self.actor.store.oauth_refresh_token_expiry = self.refresh_expiry
                    self.actor.property.oauth_refresh_token_expiry = None
            if not self.refresh_token:
                self.refresh_token = self.actor.property.oauth_refresh_token
                if self.refresh_token:
                    self.actor.store.oauth_refresh_token = self.refresh_token
                    self.actor.property.oauth_refresh_token = None
        if self.type == 'basic':
            self.realm = self.config.auth_realm
        elif self.type == 'oauth':
            if self.oauth.enabled():
                self.cookie = 'oauth_token'
                redir = self.actor.store.cookie_redirect
                if self.config.migrate_2_5_0 and not redir:
                    redir = self.actor.property.cookie_redirect
                    if redir:
                        self.actor.store.cookie_redirect = redir
                        self.actor.property.cookie_redirect = None
                if redir:
                    self.cookie_redirect = self.config.root + redir
                else:
                    self.cookie_redirect = None
                self.redirect = str(self.config.root +
                                    self.actor.id + '/oauth')
            else:
                self.type = 'none'

    def __process_oauth_accept(self, result):
        if not result:
            return None
        if not result['access_token']:
            logging.debug('No token in response')
            return None
        now = time.time()
        self.token = result['access_token']
        self.actor.store.oauth_token = self.token
        self.expiry = str(now + result['expires_in'])
        self.actor.store.oauth_token_expiry = self.expiry
        if 'refresh_token' in result:
            self.refresh_token = result['refresh_token']
            if 'refresh_token_expires_in' in result:
                self.refresh_expiry = str(
                    now + result['refresh_token_expires_in'])
            else:
                # Set a default expiry 12 months ahead
                self.refresh_expiry = str(now + (365*24*3600))
            self.actor.store.oauth_refresh_token = self.refresh_token
            self.actor.store.oauth_refresh_token_expiry = self.refresh_expiry

    def process_oauth_callback(self, code):
        """ Called when a callback is received as part of an OAuth flow to exchange code for a bearer token."""
        if not code:
            return False
        if not self.oauth:
            logging.warning(
                'Call to processOauthCallback() with oauth disabled.')
            return False
        result = self.oauth.oauth_request_token(code)
        if not result or (result and 'access_token' not in result):
            logging.warning('No token in response')
            return False
        self.__process_oauth_accept(result)
        return True

    def validate_oauth_token(self, lazy=False):
        """ Called to validate the token as part of a web-based flow.

            Returns the redirect URI to send back to the browser or empty string.
            If lazy is true, refresh_token is used only if < 24h until expiry.
        """
        if not self.token or not self.expiry:
            return self.oauth.oauth_redirect_uri(state=self.actor.id, creator=self.actor.creator)
        now = time.time()
        # Is the token still valid?
        if now < (float(self.expiry) - 20.0):
            return ""
        # Has refresh_token expired?
        if now > (float(self.refresh_expiry) - 20.0):
            return self.oauth.oauth_redirect_uri(state=self.actor.id)
        # Do we have more than a day until refresh token expiry?
        if lazy and now < (float(self.refresh_expiry) - (3600.0 * 24)):
            return ""
        # Refresh the token
        result = self.oauth.oauth_refresh_token(self.refresh_token)
        if not result:
            return self.oauth.oauth_redirect_uri(state=self.actor.id)
        self.__process_oauth_accept(result)
        return ""

    def oauth_get(self, url=None, params=None):
        """Used to call GET from the attached oauth service.

           Uses oauth.get_request(), but refreshes token if necessary.
           The function fails if token is invalid and no refresh is
           possible. For web-based flows, validate_oauth_token() needs
           to be used to validate token and get redirect URI for new
           authorization flow.
        """
        if not url:
            return None
        ret = self.oauth.get_request(url=url, params=params)
        code1 = self.oauth.last_response_code
        if (ret and any(ret)) or code1 == 204 or code1 == 201 or code1 == 404:
            return ret
        if self.actor and self.actor.id and (not ret or code1 == 401 or code1 == 403):
            refresh = self.oauth.oauth_refresh_token(
                refresh_token=self.refresh_token)
            if not refresh:
                logging.warning(
                    'Tried to refresh token and failed for Actor(' + self.actor.id + ')')
            else:
                self.__process_oauth_accept(refresh)
                ret2 = self.oauth.get_request(url=url, params=params)
                code2 = self.oauth.last_response_code
                if ret2 and any(ret2) or code2 == 204 or code2 == 201:
                    return ret2
        return None

    def oauth_head(self, url=None, params=None):
        """Used to call HEAD from the attached oauth service.

           Uses oauth.head_request((), but refreshes token if necessary.
           The function fails if token is invalid and no refresh is
           possible. For web-based flows, validate_oauth_token() needs
           to be used to validate token and get redirect URI for new
           authorization flow.
        """
        if not url:
            return None
        ret = self.oauth.head_request(url=url, params=params)
        code1 = self.oauth.last_response_code
        if (ret and any(ret)) or code1 == 204 or code1 == 201 or code1 == 404:
            return ret
        if self.actor and self.actor.id and (not ret or code1 == 401 or code1 == 403):
            refresh = self.oauth.oauth_refresh_token(
                refresh_token=self.refresh_token)
            if not refresh:
                logging.warning(
                    'Tried to refresh token and failed for Actor(' + self.actor.id + ')')
            else:
                self.__process_oauth_accept(refresh)
                ret2 = self.oauth.head_request(url=url, params=params)
                code2 = self.oauth.last_response_code
                if ret2 and any(ret2) or code2 == 204 or code2 == 201:
                    return ret2
        return None

    def oauth_delete(self, url=None):
        """Used to call DELETE from the attached oauth service.

           Uses oauth.delete_request((), but refreshes token if necessary.
           The function fails if token is invalid and no refresh is
           possible. For web-based flows, validate_oauth_token() needs
           to be used to validate token and get redirect URI for new
           authorization flow.
        """
        if not url:
            return None
        ret = self.oauth.delete_request(url=url)
        code1 = self.oauth.last_response_code
        if (ret and any(ret)) or code1 == 204 or code1 == 404:
            return ret
        if self.actor and self.actor.id and (code1 == 401 or code1 == 403):
            refresh = self.oauth.oauth_refresh_token(
                refresh_token=self.refresh_token)
            if not refresh:
                logging.warning(
                    'Tried to refresh token and failed for Actor(' + self.actor.id + ')')
            else:
                self.__process_oauth_accept(refresh)
                ret2 = self.oauth.delete_request(url=url)
                code2 = self.oauth.last_response_code
                if ret2 and any(ret2) or code2 == 204:
                    return ret2
        return None

    def oauth_post(self, url=None, params=None, urlencode=False):
        """Used to call POST from the attached oauth service.

           Uses oauth.post_request(), but refreshes token if necessary.
           The function fails if token is invalid and no refresh is
           possible. For web-based flows, validate_oauth_token() needs
           to be used to validate token and get redirect URI for new
           authorization flow.
        """
        if not url:
            return None
        ret = self.oauth.post_request(
            url=url, params=params, urlencode=urlencode)
        code1 = self.oauth.last_response_code
        if (ret and any(ret)) or code1 == 204 or code1 == 201 or code1 == 404:
            return ret
        if self.actor and self.actor.id and (code1 == 401 or code1 == 403):
            refresh = self.oauth.oauth_refresh_token(
                refresh_token=self.refresh_token)
            if not refresh:
                logging.warning(
                    'Tried to refresh token and failed for Actor(' + self.actor.id + ')')
            else:
                self.__process_oauth_accept(refresh)
                ret2 = self.oauth.post_request(
                    url=url, params=params, urlencode=urlencode)
                code2 = self.oauth.last_response_code
                if ret2 and any(ret2) or code2 == 204 or code2 == 201:
                    return ret2
        return None

    def oauth_put(self, url=None, params=None, urlencode=False):
        """Used to call PUT from the attached oauth service.

           Uses oauth.put_request(), but refreshes token if necessary.
           The function fails if token is invalid and no refresh is
           possible. For web-based flows, validate_oauth_token() needs
           to be used to validate token and get redirect URI for new
           authorization flow.
        """
        if not url:
            return None
        ret = self.oauth.put_request(
            url=url, params=params, urlencode=urlencode)
        code1 = self.oauth.last_response_code
        if (ret and any(ret)) or code1 == 204 or code1 == 201 or code1 == 404:
            return ret
        if self.actor and self.actor.id and (code1 == 401 or code1 == 403):
            refresh = self.oauth.oauth_refresh_token(
                refresh_token=self.refresh_token)
            if not refresh:
                logging.warning(
                    'Tried to refresh token and failed for Actor(' + self.actor.id + ')')
            else:
                self.__process_oauth_accept(refresh)
                ret2 = self.oauth.put_request(
                    url=url, params=params, urlencode=urlencode)
                code2 = self.oauth.last_response_code
                if ret2 and any(ret2) or code2 == 204 or code2 == 201:
                    return ret2
        return None

    # Called from a www page (browser access) to verify that a cookie has been
    # set to the actor's valid token.
    def __check_cookie_auth(self, appreq, path):
        if not path:
            path = ''
        if not self.actor:
            return False
        if self.token:
            now = time.time()
            if appreq.request.cookies and self.cookie in appreq.request.cookies:
                authz = appreq.request.cookies[self.cookie]
            else:
                authz = ''
            if appreq.request.get('refresh') and appreq.request.get('refresh').lower() == 'true':
                # Clear cookie and do a refresh if refresh=True is in GET param
                authz = ''
                self.actor.store.oauth_token = None
            elif authz == self.token and now < (float(self.expiry) - 20.0):
                logging.debug(
                    'Authorization cookie header matches a valid token')
                self.acl["relationship"] = "creator"
                self.acl["authenticated"] = True
                self.response['code'] = 200
                self.response['text'] = "Ok"
                self.authn_done = True
                return True
            elif authz != self.token:
                logging.debug(
                    'Authorization cookie header does not match a valid token')
                self.response['code'] = 403
                self.response['text'] = "Forbidden"
                self.authn_done = True
                return False
        if self.cookie_redirect:
            logging.debug('Cookie redirect already set!')
        else:
            self.actor.store.cookie_redirect = self.actor.id + path
            self.cookie_redirect = '/' + self.actor.id + path
        self.response['code'] = 302
        return False

    def set_cookie_on_cookie_redirect(self, appreq):
        """ Called after successful auth to set the cookie with the token value."""
        if not self.cookie_redirect:
            return False
        if not self.token:
            logging.warning(
                "Trying to set cookie when no token value can be found.")
            return False
        logging.debug('Setting Authorization cookie: ' + str(self.token))
        appreq.response.set_cookie(self.cookie, str(self.token),
                                   max_age=1209600, path='/', secure=True)
        appreq.response.set_redirect(str(self.cookie_redirect))
        self.actor.store.cookie_redirect = None
        return True

    def __check_basic_auth_creator(self, appreq):
        if self.type != 'basic':
            logging.warning(
                "Trying to do basic auth when auth type is not basic")
            self.response['code'] = 403
            self.response['text'] = "Forbidden"
            return False
        if not self.actor.passphrase:
            logging.warning(
                "Trying to do basic auth when no passphrase value can be found.")
            self.response['code'] = 403
            self.response['text'] = "Forbidden"
            return False
        if 'Authorization' not in appreq.request.headers:
            self.response['headers']['WWW-Authenticate'] = 'Basic realm="' + \
                self.realm + '"'
            self.response['code'] = 401
            self.response['text'] = "Authentication required"
            return False
        authz = appreq.request.headers['Authorization']
        (basic, token) = authz.split(' ')
        if basic.lower() != "basic":
            self.response['code'] = 403
            self.response['text'] = "No basic auth in Authorization header"
            logging.debug("No basic auth in Authorization header")
            return False
        self.authn_done = True
        au = authz.split(' ')[1]
        au = au.encode('utf-8')
        au = base64.b64decode(au)
        (username, password) = au.split(b':')
        password = password.decode('utf-8')
        username = username.decode('utf-8')
        if username != self.actor.creator:
            self.response['code'] = 403
            self.response['text'] = "Invalid username or password"
            logging.debug("Wrong creator username")
            return False
        if password != self.actor.passphrase:
            self.response['code'] = 403
            self.response['text'] = "Invalid username or password"
            logging.debug("Wrong creator passphrase(" +
                          password + ") correct(" + self.actor.passphrase + ")")
            return False
        self.acl["relationship"] = "creator"
        self.acl["authenticated"] = True
        self.response['code'] = 200
        self.response['text'] = "Ok"
        return True

    def check_token_auth(self, appreq):
        """ Called with an http request to check the Authorization header and validate if we have a peer with
        this token."""
        if 'Authorization' not in appreq.request.headers:
            return False
        auth = appreq.request.headers['Authorization']
        (bearer, token) = auth.split(' ')
        if bearer.lower() != "bearer":
            return False
        self.authn_done = True
        trustee = self.actor.store.trustee_root
        if self.config.migrate_2_5_0 and not trustee:
            trustee = self.actor.property.trustee_root
            if trustee:
                self.actor.property.trustee_root = None
                self.actor.store.trustee_root = trustee
        # If trustee_root is set, creator name is 'trustee' and
        # bit strength of passphrase is > 80, use passphrase as
        # token
        if trustee and self.actor.creator.lower() == 'trustee':
            if math.floor(len(self.actor.passphrase)*math.log(94, 2)) > 80:
                if token == self.actor.passphrase:
                    self.acl["relationship"] = 'trustee'
                    self.acl["peerid"] = ''
                    self.acl["approved"] = True
                    self.acl["authenticated"] = True
                    self.response['code'] = 200
                    self.response['text'] = "Ok"
                    self.token = self.actor.passphrase
                    return True
            else:
                logging.warning(
                    'Attempted trustee bearer token auth with <80 bit strength token.')
        tru = trust.Trust(actor_id=self.actor.id,
                          token=token, config=self.config)
        new_trust = tru.get()
        if new_trust:
            logging.debug('Found trust with token: (' + str(new_trust) + ')')
            if new_trust["peerid"] == self.actor.id:
                logging.error("Peer == actor!!")
                return False
        if new_trust and len(new_trust) > 0:
            self.acl["relationship"] = new_trust["relationship"]
            self.acl["peerid"] = new_trust["peerid"]
            self.acl["approved"] = new_trust["approved"]
            self.acl["authenticated"] = True
            self.response['code'] = 200
            self.response['text'] = "Ok"
            self.token = new_trust["secret"]
            self.trust = new_trust
            return True
        else:
            return False

    def check_authentication(self, appreq, path):
        """ Checks authentication in appreq, redirecting back to path if oauth is done."""
        logging.debug('Checking authentication, token auth...')
        if self.check_token_auth(appreq):
            return
        elif self.type == 'oauth':
            logging.debug('Checking authentication, oauth cookie...')
            self.__check_cookie_auth(appreq=appreq, path=path)
            return
        elif self.type == 'basic':
            logging.debug('Checking authentication, basic auth...')
            self.__check_basic_auth_creator(appreq=appreq)
            return
        logging.debug('Authentication done, and failed')
        self.authn_done = True
        self.response['code'] = 403
        self.response['text'] = "Forbidden"
        return

    def check_authorisation(self, path='', subpath='', method='', peerid='',
                            approved=True):
        """ Checks if the authenticated user has acl access rights in config.py.

        Takes the path, subpath, method, and peerid of the path (if auth user
        is different from the peer that owns the path, e.g. creator). If approved
        is False, then the trust relationship does not need to be approved for
        access"""
        if len(self.acl["peerid"]) > 0 and approved and self.acl["approved"] is False:
            logging.debug(
                'Rejected authorization because trust relationship is not approved.')
            return False
        if self.acl["relationship"]:
            relationship = self.acl["relationship"].lower()
        else:
            relationship = ''
        method = method.upper()
        self.acl["authorised"] = True
        self.acl["rights"] = "r"
        if len(path) == 0:
            return False
        if not subpath:
            subpath = ''
        fullpath = path.lower() + '/' + subpath.lower()
        # ACLs: ('role', 'path', 'METHODS', 'access')
        logging.debug('Testing access for (' + relationship +
                      ' ' + self.acl["peerid"] + ') on (' + fullpath + ' ' +
                      peerid + ') using method ' + method)
        for acl in self.config.access:
            if acl[0] == 'any' and not self.acl["authenticated"]:
                continue
            if len(acl[0]) > 0 and acl[0] != 'any' and acl[0] != relationship and acl[0] != 'owner':
                continue  # no match on relationship
            if acl[0] == relationship or acl[0] == 'any' or \
                    len(acl[0]) == 0 or (acl[0] == 'owner' and len(peerid) > 0 and
                                         self.acl["peerid"] == peerid):
                if fullpath.find(acl[1]) == 0:
                    if len(acl[2]) == 0 or acl[2].find(method) != -1:
                        self.acl["rights"] = acl[3]
                        logging.debug('Granted ' + acl[3] +
                                      ' access with ACL:' +
                                      str(acl))
                        return True
        return False



from builtins import str
import logging
import json
from actingweb import actor
from actingweb.handlers import base_handler


class RootFactoryHandler(base_handler.BaseHandler):

    def get(self):
        if self.request.get('_method') == 'POST':
            self.post()
            return
        if self.config.ui:
            self.response.template_values = {
            }
        else:
            self.response.set_status(404)

    def post(self):
        myself = actor.Actor(config=self.config)
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
            is_json = True
            if 'creator' in params:
                creator = params['creator']
            else:
                creator = ''
            if 'trustee_root' in params:
                trustee_root = params['trustee_root']
            else:
                trustee_root = ''
            if 'passphrase' in params:
                passphrase = params['passphrase']
            else:
                passphrase = ''
        except ValueError:
            is_json = False
            creator = self.request.get('creator')
            trustee_root = self.request.get('trustee_root')
            passphrase = self.request.get('passphrase')
        if not myself.create(url=self.request.url, creator=creator, passphrase=passphrase):
            self.response.set_status(400, 'Not created')
            logging.warning("Was not able to create new Actor("+str(self.request.url) + " " +
                            str(creator) + ")")
            return
        if len(trustee_root) > 0:
            myself.store.trustee_root = trustee_root
        self.response.headers["Location"] = str(self.config.root + myself.id)
        if self.config.www_auth == 'oauth' and not is_json:
            self.response.set_redirect(self.config.root + myself.id + '/www')
            return
        pair = {
            'id': myself.id,
            'creator': myself.creator,
            'passphrase': str(myself.passphrase),
        }
        if len(trustee_root) > 0:
            pair['trustee_root'] = trustee_root
        if self.config.ui and not is_json:
            self.response.template_values = pair
            return
        out = json.dumps(pair)
        self.response.write(out)
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(201, 'Created')



from builtins import str
import json
import logging
from actingweb import auth
from actingweb.handlers import base_handler


class SubscriptionRootHandler(base_handler.BaseHandler):
    """Handles requests to /subscription"""

    def get(self, actor_id):
        if self.request.get('_method') == 'POST':
            self.post(actor_id)
            return
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='subscriptions',
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='subscriptions', method='GET'):
            self.response.set_status(403)
            return
        peerid = self.request.get('peerid')
        target = self.request.get('target')
        subtarget = self.request.get('subtarget')
        resource = self.request.get('resource')

        subscriptions = myself.get_subscriptions(peerid=peerid, target=target, subtarget=subtarget, resource=resource)
        if not subscriptions:
            self.response.set_status(404, 'Not found')
            return
        data = {
                'id': myself.id,
                'data': subscriptions,
                }
        out = json.dumps(data)
        self.response.write(out)
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(200, 'Ok')

    def post(self, actor_id):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='subscriptions',
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='subscriptions', method='POST'):
            self.response.set_status(403)
            return
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
            if 'peerid' in params:
                peerid = params['peerid']
            else:
                peerid = None
            if 'target' in params:
                target = params['target']
            else:
                target = None
            if 'subtarget' in params:
                subtarget = params['subtarget']
            else:
                subtarget = None
            if 'resource' in params:
                resource = params['resource']
            else:
                resource = None
            if 'granularity' in params:
                granularity = params['granularity']
            else:
                granularity = 'none'
        except ValueError:
            peerid = self.request.get('peerid')
            target = self.request.get('target')
            subtarget = self.request.get('subtarget')
            resource = self.request.get('resource')
            granularity = self.request.get('granularity')
        if not peerid or len(peerid) == 0:
            self.response.set_status(400, 'Missing peer URL')
            return
        if not target or len(target) == 0:
            self.response.set_status(400, 'Missing target')
            return
        remote_loc = myself.create_remote_subscription(
            peerid=peerid, target=target, subtarget=subtarget, resource=resource, granularity=granularity)
        if not remote_loc:
            self.response.set_status(408, 'Unable to create remote subscription with peer')
            return
        self.response.headers["Location"] = remote_loc
        self.response.set_status(204, 'Created')


# Handling requests to /subscription/*, e.g. /subscription/<peerid>
class SubscriptionRelationshipHandler(base_handler.BaseHandler):

    def get(self, actor_id, peerid):
        if self.request.get('_method') == 'POST':
            self.post(actor_id, peerid)
            return
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='subscriptions',
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='subscriptions', method='GET', peerid=peerid):
            self.response.set_status(403)
            return
        target = self.request.get('target')
        subtarget = self.request.get('subtarget')
        resource = self.request.get('resource')

        subscriptions = myself.get_subscriptions(peerid=peerid, target=target, subtarget=subtarget, resource=resource)
        if not subscriptions:
            self.response.set_status(404, 'Not found')
            return
        data = {
                'id': myself.id,
                'peerid': peerid,
                'data': subscriptions,
                }
        out = json.dumps(data)
        self.response.write(out)
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(200, 'Ok')

    def post(self, actor_id, peerid):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='subscriptions',
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='subscriptions', subpath='<id>', method='POST', peerid=peerid):
            self.response.set_status(403)
            return
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
            if 'target' in params:
                target = params['target']
            else:
                self.response.set_status(400, 'No target in request')
                return
            if 'subtarget' in params:
                subtarget = params['subtarget']
            else:
                subtarget = None
            if 'resource' in params:
                resource = params['resource']
            else:
                resource = None
            if 'granularity' in params:
                granularity = params['granularity']
            else:
                granularity = 'none'
        except ValueError:
            self.response.set_status(400, 'No json body')
            return
        if peerid != check.acl["peerid"]:
            logging.warning("Peer " + peerid +
                         " tried to create a subscription for peer " + check.acl["peerid"])
            self.response.set_status(403, 'Forbidden. Wrong peer id in request')
            return
        # We need to validate that this peer has GET rights on what it wants to subscribe to
        if not check.check_authorisation(path=target, subpath=subtarget, method='GET', peerid=peerid):
            self.response.set_status(403)
            return
        new_sub = myself.create_subscription(
            peerid=check.acl["peerid"], target=target, subtarget=subtarget, resource=resource, granularity=granularity)
        if not new_sub:
            self.response.set_status(500, 'Unable to create new subscription')
            return
        self.response.headers["Location"] = str(self.config.root + 
                                                myself.id + '/subscriptions/' + new_sub["peerid"] +
                                                '/' + new_sub["subscriptionid"])
        pair = {
            'subscriptionid': new_sub["subscriptionid"],
            'target': new_sub["target"],
            'subtarget': new_sub["subtarget"],
            'resource': new_sub["resource"],
            'granularity': new_sub["granularity"],
            'sequence': new_sub["sequence"],
        }
        out = json.dumps(pair)
        self.response.write(out)
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(201, 'Created')

    
class SubscriptionHandler(base_handler.BaseHandler):
    """ Handling requests to specific subscriptions, e.g. /subscriptions/<peerid>/12f2ae53bd"""

    def get(self, actor_id, peerid, subid):
        if self.request.get('_method') == 'PUT':
            self.put(actor_id, peerid, subid)
            return
        if self.request.get('_method') == 'DELETE':
            self.delete(actor_id, peerid, subid)
            return
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='subscriptions',
                                              subpath=peerid + '/' + subid,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='subscriptions', subpath='<id>/<id>', method='GET', peerid=peerid):
            self.response.set_status(403)
            return
        sub = myself.get_subscription_obj(peerid=peerid, subid=subid)
        sub_data = sub.get()
        if not sub_data or len(sub_data) == 0:
            self.response.set_status(404, "Subscription does not exist")
            return
        diffs = sub.get_diffs()
        pairs = []
        for diff in diffs:
            try:
                d = json.loads(diff["diff"])
            except (TypeError, ValueError, KeyError):
                d = diff["diff"]
            pairs.append({
                'sequence': diff["sequence"],
                'timestamp': diff["timestamp"].strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
                'data': d,
            })
        if len(pairs) == 0:
            self.response.set_status(404, 'No diffs available')
            return
        data = {
                'id': myself.id,
                'peerid': peerid,
                'subscriptionid': subid,
                'target': sub_data["target"],
                'subtarget': sub_data["subtarget"],
                'resource': sub_data["resource"],
                'data': pairs,
                }
        out = json.dumps(data)
        self.response.write(out)
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(200, 'Ok')

    def put(self, actor_id, peerid, subid):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='subscriptions',
                                              subpath=peerid + '/' + subid,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='subscriptions', subpath='<id>/<id>', method='GET', peerid=peerid):
            self.response.set_status(403)
            return
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
            if 'sequence' in params:
                seq = params['sequence']
            else:
                self.response.set_status(400, "Error in json body and no GET parameters")
                return
        except (TypeError, ValueError, KeyError):
            seq = self.request.get('sequence')
            if len(seq) == 0:
                self.response.set_status(400, "Error in json body and no GET parameters")
                return
        try:
            if not isinstance(seq, int):
                seqnr = int(seq)
            else:
                seqnr = seq
        except ValueError:
            self.response.set_status(400, "Sequence does not contain a number")
            return
        sub = myself.get_subscription_obj(peerid=peerid, subid=subid)
        if not sub:
            self.response.set_status(404, "Subscription does not exist")
            return
        sub.clear_diffs(seqnr=seqnr)
        self.response.set_status(204)
        return

    def delete(self, actor_id, peerid, subid):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='subscriptions',
                                              subpath=peerid + '/' + subid,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='subscriptions', subpath='<id>/<id>', method='GET', peerid=peerid):
            self.response.set_status(403)
            return
        # Do not delete remote subscription if this is from our peer
        if len(check.acl['peerid']) == 0:
            myself.delete_remote_subscription(peerid=peerid, subid=subid)
        if not myself.delete_subscription(peerid=peerid, subid=subid):
            self.response.set_status(404)
            return
        self.response.set_status(204)
        return
        

class SubscriptionDiffHandler(base_handler.BaseHandler):
    """ Handling requests to specific diffs for one subscription and clears it, e.g. 
    /subscriptions/<peerid>/<subid>/112"""

    def get(self, actor_id, peerid, subid, seqnr):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='subscriptions',
                                              subpath=peerid + '/' + subid + '/' + str(seqnr),
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='subscriptions', subpath='<id>/<id>', method='GET', peerid=peerid):
            self.response.set_status(403)
            return
        sub = myself.get_subscription_obj(peerid=peerid, subid=subid)
        sub_data = sub.get()
        if not sub:
            self.response.set_status(404, "Subscription does not exist")
            return
        if not isinstance(seqnr, int):
            seqnr = int(seqnr)
        diff = sub.get_diff(seqnr=seqnr)
        if not diff:
            self.response.set_status(404, 'No diffs available')
            return
        try:
            d = json.loads(diff["data"])
        except (TypeError, ValueError, KeyError):
            d = diff["data"]
        pairs = {
            'id': myself.id,
            'peerid': peerid,
            'subscriptionid': subid,
            'timestamp': diff["timestamp"].strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
            'target': sub_data["target"],
            'subtarget': sub_data["subtarget"],
            'resource': sub_data["resource"],
            'sequence': seqnr,
            'data': d,
        }
        sub.clear_diff(seqnr)
        out = json.dumps(pairs)
        self.response.write(out)
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(200, 'Ok')



__all__ = [
    "root",
    "bot",
    "callbacks",
    "callback_oauth",
    "devtest",
    "meta",
    "oauth",
    "properties",
    "resources",
    "factory",
    "subscription",
    "trust",
    "www",
    "base_handler",
]



from builtins import object
import os
import logging
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute

"""
    DbPeerTrustee handles all db operations for a peer we are a trustee for.
    Google datastore for google is used as a backend.
"""


class PeerTrustee(Model):

    class Meta(object):
        table_name = os.getenv('AWS_DB_PREFIX', 'demo_actingweb') + "_peertrustees"
        read_capacity_units = 1
        write_capacity_units = 1
        region = os.getenv('AWS_DEFAULT_REGION', 'us-west-1')
        host = os.getenv('AWS_DB_HOST', None)

    id = UnicodeAttribute(hash_key=True)
    peerid = UnicodeAttribute(range_key=True)
    baseuri = UnicodeAttribute()
    type = UnicodeAttribute()
    passphrase = UnicodeAttribute()


class DbPeerTrustee(object):
    """
        DbPeerTrustee does all the db operations for property objects

        The actor_id must always be set.
    """

    def get(self,  actor_id=None, peer_type=None, peerid=None):
        """ Retrieves the peertrustee from the database """
        if not actor_id:
            return None
        if not peerid and not peer_type:
            logging.debug("Attempt to get DbPeerTrustee without peerid or type")
            return None
        try:
            if not self.handle and peerid:
                self.handle = PeerTrustee.get(actor_id, peerid, consistent_read=True)
            elif not self.handle and peer_type:
                count = 0
                peer_trustees = PeerTrustee.scan((PeerTrustee.id == actor_id) & (PeerTrustee.type == peer_type))
                for h in peer_trustees:
                    self.handle = h
                    count+=1
                if count > 1:
                    logging.error('Found more than one peer of this peer trustee type(' +
                                  peer_type + '). Unable to determine which, need peerid lookup.')
                    return False
                if count == 0:
                    self.handle = None
        except PeerTrustee.DoesNotExist:
            self.handle = None
        if self.handle:
            t = self.handle
            return {
                "id": t.id,
                "peerid": t.peerid,
                "baseuri": t.baseuri,
                "type": t.type,
                "passphrase": t.passphrase,
            }
        else:
            return None

    def create(self, actor_id=None,
               peerid=None,
               peer_type=None,
               baseuri=None,
               passphrase=None):
        """ Create a new peertrustee """
        if not actor_id or not peerid or not peer_type:
            logging.debug("actor_id, peerid, and type are mandatory when creating peertrustee in db")
            return False
        if not baseuri:
            baseuri = ''
        if not passphrase:
            passphrase = ''
        self.handle = PeerTrustee(id=actor_id,
                                  peerid=peerid,
                                  type=peer_type,
                                  baseuri=baseuri,
                                  passphrase=passphrase)
        self.handle.save()
        return True

    def modify(self,
               peer_type=None,
               baseuri=None,
               passphrase=None):
        """ Modify a peertrustee

            If bools are none, they will not be changed.
        """
        if not self.handle:
            logging.debug("Attempted modification of DbPeerTrustee without db handle")
            return False
        if baseuri and len(baseuri) > 0:
            self.handle.baseuri = baseuri
        if passphrase and len(passphrase) > 0:
            self.handle.passphrase = passphrase
        if type and len(peer_type) > 0:
            self.handle.type = peer_type
        self.handle.save()
        return True

    def delete(self):
        """ Deletes the peertrustee in the database after a get() """
        if not self.handle:
            return False
        self.handle.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        if not PeerTrustee.exists():
            PeerTrustee.create_table(wait=True)


class DbPeerTrusteeList(object):
    """
        DbPeerTrusteeList does all the db operations for list of peertrustee objects

        The  actor_id must always be set.
    """

    def fetch(self,  actor_id=None):
        """ Retrieves the peer trustees of an actor_id from the database """
        if not actor_id:
            return None
        self.actor_id = actor_id
        self.peertrustees = []
        self.handle = PeerTrustee.scan(PeerTrustee.actor_id == self.actor_id)
        if self.handle:
            for t in self.handle:
                self.peertrustees.append({
                  "id": t.id,
                  "peerid": t.peerid,
                  "baseuri": t.baseuri,
                  "type": t.type,
                  "passphrase": t.passphrase,
                })
            return self.peertrustees
        else:
            return []

    def delete(self):
        """ Deletes all the peertrustees in the database """
        self.handle = PeerTrustee.scan(PeerTrustee.actor_id == self.actor_id)
        if not self.handle:
            return False
        for p in self.handle:
            p.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        self.actor_id = None
        self.peertrustees = None



from builtins import object


class AWRequest(object):

    def get_header(self, header=''):
        header = header.lower()
        for k, v in self.headers.items():
            if header == k.lower():
                return v
        return ''

    def get(self, var=''):
        var = var.lower()
        for k, v in self.params.items():
            if var == k.lower():
                return v
        return ''

    def arguments(self):
        ret = []
        for k, v in self.params.items():
            ret.append(k)
        return ret

    def __init__(self, url=None, params=None, body=None, headers=None, cookies=None):
        self.headers = headers
        self.params = params
        self.body = body
        self.url = url
        self.cookies = cookies


class AWResponse(object):

    def set_status(self, code=200, message='Ok'):
        if not code or code < 100 or code > 599:
            return False
        if not message:
            message = ''
        self.status_code = code
        self.status_message = message

    def write(self, body=None, encode=False):
        if not body:
            return False
        if encode:
            self.body = body.encode('utf-8')
        else:
            self.body = body

    def set_cookie(self, name, value, max_age=1209600, path='/', secure=True):
        self.cookies.append({
            'name': name,
            'value': value,
            'max_age': max_age,
            'path': path,
            'secure': secure
        })

    def set_redirect(self, url):
        self.redirect = url

    def __init__(self):
        self.status_code = 200
        self.status_message = 'Ok'
        self.headers = {}
        self.body = ''
        self.redirect = None
        self.cookies = []
        self.template_values = {}


class AWWebObj(object):

    def __init__(self, url=None, params=None, body=None, headers=None, cookies=None):
        self.request = AWRequest(url=url, params=params, body=body, headers=headers, cookies=cookies)
        self.response = AWResponse()



from builtins import str
import json
import datetime
from actingweb import auth
from actingweb import aw_proxy
from actingweb import attribute
from actingweb.handlers import base_handler


class DevtestHandler(base_handler.BaseHandler):

    def put(self, actor_id, path):
        """Handles PUT for devtest"""

        if not self.config.devtest:
            self.response.set_status(404)
            return
        (myself, check) = auth.init_actingweb(
            appreq=self,
            actor_id=actor_id,
            path='devtest',
            subpath=path,
            config=self.config)
        if not myself or check.response["code"] != 200:
            return
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
        except (TypeError, ValueError, KeyError):
            params = self.request.body.decode('utf-8', 'ignore')
        paths = path.split('/')
        if paths[0] == 'proxy':
            mytwin = myself.get_peer_trustee(shorttype='myself')
            if mytwin and len(mytwin) > 0:
                if paths[1] == 'properties' and paths[2] and len(paths[2]) > 0:
                        proxy = aw_proxy.AwProxy(peer_target=mytwin, config=self.config)
                        if params:
                            proxy.change_resource('/properties/' + paths[2], params=params)
                        self.response.set_status(proxy.last_response_code)
                        return
        elif paths[0] == 'ping':
            self.response.set_status(204)
            return
        elif paths[0] == 'attribute':
            if len(paths) > 2:
                bucket = attribute.Attributes(actor_id=myself.id, bucket=paths[1], config=self.config)
                bucket.set_attr(paths[2], params, timestamp=datetime.datetime.utcnow())
                self.response.set_status(204)
                return
        self.response.set_status(404)

    def delete(self, actor_id, path):
        """Handles DELETE for devtest"""

        if not self.config.devtest:
            self.response.set_status(404)
            return
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id, path='devtest',
                                              subpath=path,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        paths = path.split('/')
        if paths[0] == 'proxy':
            mytwin = myself.get_peer_trustee(shorttype='myself')
            if mytwin and len(mytwin) > 0:
                if paths[1] == 'properties':
                    proxy = aw_proxy.AwProxy(peer_target=mytwin, config=self.config)
                    proxy.delete_resource(path='/properties')
                    self.response.set_status(proxy.last_response_code)
                    return
        elif paths[0] == 'ping':
            self.response.set_status(204)
            return
        elif paths[0] == 'attribute':
            if len(paths) > 2:
                bucket = attribute.Attributes(actor_id=myself.id, bucket=paths[1], config=self.config)
                bucket.delete_attr(paths[2])
                self.response.set_status(204)
                return
            else:
                buckets = attribute.Buckets(actor_id=myself.id, config=self.config)
                buckets.delete()
                self.response.set_status(204)
                return
        self.response.set_status(404)

    def get(self, actor_id, path):
        """Handles GET for devtest"""

        if not self.config.devtest:
            self.response.set_status(404)
            return
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id, path='devtest',
                                              subpath=path,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        paths = path.split('/')
        if paths[0] == 'proxy':
            mytwin = myself.get_peer_trustee(shorttype='myself')
            if mytwin and len(mytwin) > 0:
                if paths[1] == 'properties':
                    proxy = aw_proxy.AwProxy(peer_target=mytwin, config=self.config)
                    prop = proxy.get_resource(path='/properties')
                    if proxy.last_response_code != 200:
                        self.response.set_status(proxy.last_response_code)
                        return
                    out = json.dumps(prop)
                    self.response.write(out.encode('utf-8'))
                    self.response.headers["Content-Type"] = "application/json"
                    self.response.set_status(200)
                    return
        elif paths[0] == 'ping':
            self.response.set_status(204)
            return
        elif paths[0] == 'attribute':
            if len(paths) > 1:
                bucket = attribute.Attributes(actor_id=myself.id, bucket=paths[1], config=self.config)
                params = bucket.get_bucket()
                for k, v in params.items():
                    params[k]["timestamp"] = v["timestamp"].strftime("%Y-%m-%d %H:%M:%S")
                out = json.dumps(params)
                self.response.write(out.encode('utf-8'))
                self.response.headers["Content-Type"] = "application/json"
                self.response.set_status(200)
                return
            else:
                buckets = attribute.Buckets(actor_id=myself.id, config=self.config)
                params = buckets.fetch()
                if len(params) == 0:
                    self.response.set_status(404)
                    return
                for b, d in params.items():
                    for k, v in d.items():
                        d[k]["timestamp"] = v["timestamp"].strftime("%Y-%m-%d %H:%M:%S")
                out = json.dumps(params)
                self.response.write(out.encode('utf-8'))
                self.response.headers["Content-Type"] = "application/json"
                self.response.set_status(200)
                return
        self.response.set_status(404)

    def post(self, actor_id, path):
        """Handles POST for devtest"""

        if not self.config.devtest:
            self.response.set_status(404)
            return
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id, path='devtest',
                                              subpath=path,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
        except (TypeError, ValueError, KeyError):
            params = None
        paths = path.split('/')
        if paths[0] == 'proxy':
            mytwin = myself.get_peer_trustee(shorttype='myself')
            if mytwin and len(mytwin) > 0:
                if paths[1] == 'create':
                        proxy = aw_proxy.AwProxy(peer_target=mytwin, config=self.config)
                        meta = proxy.get_resource(path='/meta')
                        if params:
                            proxy.create_resource('/properties', params=params)
                        out = json.dumps(meta)
                        self.response.write(out.encode('utf-8'))
                        self.response.headers["Content-Type"] = "application/json"
                        self.response.headers["Location"] = str(mytwin["baseuri"])
                        self.response.set_status(200)
                        return
        elif paths[0] == 'ping':
            self.response.set_status(204)
            return
        elif paths[0] == 'attribute':
            if paths[1] and len(paths[1]) > 0:
                bucket = attribute.Attributes(actor_id=myself.id, bucket=paths[1], config=self.config)
                for k, v in params.items():
                    bucket.set_attr(k, v, timestamp=datetime.datetime.utcnow())
                out = json.dumps(params)
                self.response.write(out.encode('utf-8'))
                self.response.headers["Content-Type"] = "application/json"
                self.response.set_status(200)
                return
        self.response.set_status(404)
